module Optimizer where

import Instruction
import Coder

optimize = resolvFlagReset . concat . optimizeBlocks . splitBlocks
requiredOptimize = resolvFlagReset

optimizeBlocks = map (optimizeAMMABlock.optimizeTiyBlock)

optimizeTiyBlock block@((Label _, _):is) = block
optimizeTiyBlock block = optimizeTiyBlock' [] [] block

optimizeTiyBlock' _ result [] = result
optimizeTiyBlock' [c] result (i@(OperandInt1 "TIY" v, comment):is)
    | c == v =
        if any (\x -> x) (map (isBreakRegisterY.fst) result) then
           result ++ [i] ++  optimizeTiyBlock' [v] []  is
        else
           result ++ optimizeTiyBlock' [v] [] is
optimizeTiyBlock' cc result (i@(OperandInt1 "TIY" v, comment):is) =
    result ++ [i] ++ optimizeTiyBlock' [v] [] is
optimizeTiyBlock' cc result (i:is) =
    optimizeTiyBlock' cc (result ++ [i]) is


optimizeAMMABlock [] = []
optimizeAMMABlock ((i@(Operand0 "AM", comment1)):(Operand0 "MA", comment2):is) =
    (Operand0 "AM", comment1 ++ " OOMA " ++ comment2) : optimizeAMMABlock is
optimizeAMMABlock (i:is) = i : optimizeAMMABlock is

splitBlocks [] = []
splitBlocks is =
    case s of
      [] -> [f]
      _ -> f : [head s] : splitBlocks (tail s)
    where
      b = break (\x -> isLabel (fst x)) is
      f = fst b
      s = snd b

showBlock :: [(Instruction, String)] -> String
showBlock = unlines.(map codingMnemonic)
showBlocks = unlines . map (++(take 60 (repeat '-'))) . (map showBlock)

resolvFlagReset = resolvFlagReset' [] []
resolvFlagReset' [c] ps [] = c: ps
resolvFlagReset' [] ps [] = ps
resolvFlagReset' [(ci, ccomment)] ps ((ri, rcomment):is)
    | isResetFlag ci && (ri == FlagResetDontCare || ri == FlagResetRegistersSafe) = 
        resolvFlagReset' [(ci, ccomment ++ " OORF")] ps is
resolvFlagReset' cc ps ((FlagResetDontCare, rcomment):is) =
    cc ++ ps ++ [(Operand0 "CH", "reset flag")] ++ resolvFlagReset' [] [] is
resolvFlagReset' cc ps ((FlagResetRegistersSafe, rcomment):is) =
    cc ++ ps ++ [(Operand0 "CH", "reset flag safe"), (Operand0 "CH", "")] ++ resolvFlagReset' [] [] is
resolvFlagReset' cc ps (i:is) | (not.isInstruction) (fst i) =
    resolvFlagReset' cc (ps++[i]) is
resolvFlagReset' cc ps (rr:is) =
    cc ++ ps ++ resolvFlagReset' [rr] [] is
